#!/usr/bin/env python
# coding: utf-8

# In[ ]:


### Curso de Python para Finanças Quantitativas

#### Aula 11 - Regressão Linear e Volatilidade
#### Autor: Leandro Guerra - Outspoken Market
#### Download em: https://www.outspokenmarket.com/pythonfinancasquantitativas.html


# Como pré requisito para esta aula, assista o Aulão de Regressão Linear para Leigos
# https://www.youtube.com/watch?v=0mFvllux_zQ

# ____________________________________________________

# In[1]:


# Carregando as bibliotecas

import warnings
warnings.filterwarnings("ignore")

import pandas as pd 
import numpy as np 
import matplotlib.pyplot as plt 
import seaborn as sns
from sklearn.model_selection import train_test_split 
#from sklearn.linear_model import LinearRegression 
from sklearn import metrics 
import statsmodels.api as sm
from statsmodels.formula.api import ols

import yfinance as yf

get_ipython().run_line_magic('matplotlib', 'inline')


# In[2]:


# Carregando a base de dados e fazendo os checks iniciais

plt.style.use("fivethirtyeight")
# https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html

# Carrega a base
ticker = "PETR4.SA"
df1 = yf.download(ticker, "2010-01-01", "2022-12-31")

with plt.style.context("seaborn-deep"):
    plt.figure(figsize = (10, 8))
    plt.plot(df1["Adj Close"], linewidth = 1.5)
    plt.xlabel("Ano")
    plt.ylabel("Preço")
    plt.title("Cotação Histórica " + ticker)


# In[3]:


# Cria variáveis: Volatilidade, Distância e Direção e o alvo que será a volatilidade de 5 dias

periodos = 5

df1["Retorno"] = df1["Adj Close"].pct_change(1)

df1["Retorno_p"] = df1["Adj Close"].pct_change(periodos).rolling(10).mean()

# Proporção do corpo do candle em relacão ao range do dia
df1["Prop"] = (df1["Adj Close"]-df1["Open"])/(df1["High"]-df1["Low"])

# Volatilidade de 5 dias
df1["Std"] = df1["Retorno"].rolling(5).std()

# Construção do alvo
# Alvo - Volatilidade
df1["Alvo"] = df1["Std"].shift(-periodos)


# In[5]:


df1.head(15)


# In[6]:


# Histograma

config = dict(histtype = "stepfilled", alpha = 0.6, density = False, bins = 50)

plt.figure(figsize = (10,7))
plt.hist(df1["Alvo"], **config, label = "Vol 5 dias", color = "blue")

# Formatação da legenda - Válida para qualquer tipo de gráfico
plt.legend(loc = "upper right"
           , frameon = True
           , ncol = 2 
           , fancybox = True
           , framealpha = 0.95
           , shadow = True
           , borderpad = 1)
plt.title("Distribuição da Volatilidade");


# ____________________________________________________

# ### Analisando as correlações

# In[7]:


# Analisando a correlação das variáveis com o nosso alvo

correl = df1.corr()["Alvo"].drop(["Open","High", "Low", "Close", "Adj Close", "Alvo"])

correl


# In[9]:


# Visualizando as correlações - Geralmemte heat maps são o melhor tipo de visualização

correl_hm = df1.drop(["Open","High", "Low", "Close", "Adj Close"], axis = 1).corr()

fig, ax = plt.subplots(figsize = (7, 7))

sns.heatmap(correl_hm, cmap = "coolwarm", annot = True, linewidths = 0.6, ax = ax);


# In[10]:


df1.drop(["Open","High", "Low", "Close", "Adj Close"], axis = 1).describe()


# ____________________________________________________

# ### Treinamento e teste da regressao linear com Statsmodels

# In[11]:


# Filtrando os valores missing

df1.replace([np.inf, -np.inf], np.nan, inplace = True)
df1_filtrado = df1.dropna(axis = 0) 

df1_filtrado.columns


# In[13]:


# Separando os dados entre treinamento e teste

# Vamos treinar o modelo de 2010 a 2015
start_train = "2010-01-01"
end_train = "2015-12-31"

# Vamos testar o modelo de 2016 a 2022
start_test = "2016-01-01"
end_test = "2022-12-31"

df1_train = df1_filtrado.loc[start_train : end_train]

df1_test = df1_filtrado.loc[start_test : end_test]

# Separando os dados com as variaveis em x e o alvo em y
 
x_train = df1_train.iloc[:, 5:10]
y_train = df1_train["Alvo"]

x_test = df1_test.iloc[:, 5:10]
y_test = df1_test["Alvo"]


# In[14]:


# Treinando o modelo com statsmodels

# Nota importante: por padrão o statsmodels não acrescenta o intercepto
# você deve acrescentá-lo manualmente

x_train_ = sm.add_constant(x_train)
x_test_ = sm.add_constant(x_test)

lr_sm = sm.OLS(y_train, x_train_).fit() # Linha que treina o modelo

y_pred_train = lr_sm.predict(x_train_)

y_pred_test = lr_sm.predict(x_test_)

# Mostra as estatísticas 
lr_sm.summary()


# ____________________________________________________

# ### Visualização dos resultados

# In[15]:


# Resultados: Real x previsto em um dataframe

resultados = pd.DataFrame({'Real': y_test, 'Previsto': y_pred_test})

# Isto será usado para simular a medição da acurácia do modelo
resultados["Real_lag"] = resultados["Real"].shift(1)
resultados["Previsto_lag"] = resultados["Previsto"].shift(1)
resultados["Real_dir"] = np.where(resultados["Real"] > resultados["Real_lag"], 1, 0)
resultados["Previsto_dir"] = np.where(resultados["Previsto"] > resultados["Previsto_lag"], 1, 0)
resultados["Real_cat"] = np.where(resultados["Real"] > 0.024408, 1, 0)
resultados["Previsto_cat"] = np.where(resultados["Previsto"] > 0.024408, 1, 0)

resultados


# In[16]:


# Gráfico dos resultados

plt.figure(figsize = (8, 8))
plt.plot(y_test, y_pred_test, "x", color = "blue", label = "Real x Previsto")
plt.legend()
plt.xlabel("Volatilidade prevista")
plt.ylabel("Volatilidade real")
plt.title("Volatilidade " + ticker + " - Regressão Linear");


# In[17]:


with plt.style.context("seaborn-deep"):
    plt.figure(figsize = (10, 10))
    plt.plot(resultados["Real"], linewidth = 1, label = "Real", c = "red")
    plt.plot(resultados["Previsto"], "--", linewidth = 1, label = "Prevista", c = "blue")
    plt.xlabel("Ano")
    plt.ylabel("Volatilidade")
    plt.legend(loc = "upper left"
           , frameon = True
           , fancybox = True
           , framealpha = 1
           , borderpad = 1)
    plt.title("Volatilidade " + ticker + " - Regressão Linear")


# ### Métricas de avaliação: MAE, R2 e RMSE

# In[20]:


# Avaliaçao dos resultados com statsmodels

MAE_train_sm = sm.tools.eval_measures.meanabs(y_train, y_pred_train)
R2_train_sm = lr_sm.rsquared # Coeficiente de determinaçao
RMSE_train_sm = sm.tools.eval_measures.rmse(y_train, y_pred_train)

print("----- Avaliação do treinamento -----")
print("MAE: ", round(MAE_train_sm, 3))
print('R2: ', round(R2_train_sm, 3))
print('RMSE: ', round(RMSE_train_sm, 3))

MAE_test_sm = sm.tools.eval_measures.meanabs(y_test, y_pred_test) 
RMSE_test_sm = sm.tools.eval_measures.rmse(y_test, y_pred_test)

print("")
print("----- Avaliação do teste -----")
print("MAE: ", round(MAE_test_sm, 3))
print('RMSE: ', round(RMSE_test_sm, 3))


# In[22]:


# Olhando apenas para a base de teste

print("A volatilidade média da base é: ")
print(round(y_test.mean(), 3))
print()
print("O percentual do MAE em relaçao à média da base: ")
print(round(metrics.mean_absolute_error(y_test, y_pred_test)/y_test.mean()*100, 2))


# In[23]:


# Porém...

round(pd.crosstab(resultados.Previsto_dir, resultados.Real_dir
            ,  normalize = True)*100, 2)


# In[26]:


# Porém 2...muito melhor - soma a diagnoal principal

round(pd.crosstab(resultados.Previsto_cat, resultados.Real_cat
            , normalize = True)*100, 2)


# In[28]:


# Mostrando os coeficientes (pesos) da regressao

print(lr_sm.params)


# ### A formula da sua regressão é
# 
# lr = 1.761499e-02 + volume*7.649406e-11 - 2.411297e-02*retorno e assim por diante

# ____________________________________________________

# ### Diagnostico da regressão

# In[29]:


lr_sm.resid 


# In[30]:


# Residuals x Fitted Values

plt.figure(figsize = (8, 5))
plt.plot(lr_sm.resid, y_pred_train, ".", color = "blue")
plt.xlabel("Residuals")
plt.ylabel("Fitted Values")
plt.title("Volatilidade - Residuals x Fitted Values");


# In[31]:


# Normality plot

import scipy as sp

fig, ax = plt.subplots(figsize = (6 , 6))

sp.stats.probplot(lr_sm.resid, plot = ax, fit = True);


# In[32]:


import statsmodels.api as sm

fig = plt.figure(figsize = (12, 12))

fig = sm.graphics.plot_partregress_grid(lr_sm, fig = fig)


# In[37]:


# Fazendo a previsão de um único registro

lr_sm.predict([[7, 0.010, 0.002, 0.076, 10.0, 0.012]])


# In[33]:


df1_filtrado.columns


# ____________________________________________________
